/***********************************************************************\
*                               console.d                               *
*                                                                       *
*                       Stewart's Utility Library                       *
*                                                                       *
*                              Version 0.03                             *
*                                                                       *
*      Console input/output with support for Windows character sets     *
>***********************************************************************<
* Copyright (C) 2006 Stewart Gordon.                                    *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For updates, documentation and other information related to this      *
* product, visit http://pr.stewartsplace.org.uk/d/sutil/.               *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.libs.util.console;

private import std.string, std.utf, std.format;
private import std.c.stdio, std.c.windows.windows;

/*!	\file		console.d
 *	\brief		Console input and output
 */

/*!	\package	smjg.libs.util.console
 *	\brief		Console input and output
 *
 *	The smjg.libs.util.console module provides a means of communicating
 *	with the standard input and output streams.  It is designed to
 *	provide the following benefits over stream classes in Phobos:
 *	\li support Windows character sets
 *	\li fix a bug in Win9x whereby a line of output is lost on stdin EOF
 *	\li flag EOF of stdin immediately from the class user's point of view.
 *
 *	The application programmer would use this module by means of the cIn,
 *	cOut and cErr objects.
 *
 *	This is not intended to be a sophisticated text I/O library.  The
 *	TextStream class in Indigo will provide something much more powerful
 *	when and if it is completed.
 *
 *	Any application that uses this module should use it consistently
 *	throughout if possible.  This ensures that the EOF fix works correctly
 *	and prevents buffering conflicts between different I/O libraries/APIs.
 *
 *	\since  0.03
 *	\todo   Implement error checking.  (Not sure how much is needed.)
 */


version (Windows) {
	private import std.windows.charset;
	extern (C) int isatty(int);
}


//!	The ConsoleInput class is used to receive text from the standard
//!	input.
/*!
 *	This class buffers input on a line-by-line basis.
 *
 *	\since  0.03
 */
class ConsoleInput {
	//!	Reads a single Unicode character from the input stream.
	/*!
	 *	\return  the character retrieved, or dchar.init on EOF.
	 */
	dchar readChar() {
		if (eof) return dchar.init;

		return decode(buffer, bufPos);
	}

	//!	Discards the currently buffered line of input.
	/*!
	 *	The next call to readChar, readLine or eof will therefore wait for
	 *	a line of input from the user and use it.
	 *
	 *	If no text is waiting in the buffer, this function has no effect.
	 */
	void discardLine() {
		bufPos = buffer.length = 0;
	}

	//!	Returns the number of bytes currently waiting in the internal buffer.
	/*!
	 *	This can be used to determine whether there is any more text to read
	 *	before waiting for the user to input more.
	 *
	 *	\return  the number of bytes of UTF-8 text in the buffer, including
	 *	         the terminating newline if present.
	 *	\sa      bytesInLine, charsInBuffer, charsInLine
	 */
	size_t bytesInBuffer() {
		return buffer.length - bufPos;
	}

	//!	Returns the number of bytes remaining on the current line of input.
	/*!
	 *	This can be used to determine whether there is any text, other than
	 *	a newline, which can be read before waiting for the user to input
	 *	more.
	 *
	 *	\return  the number of bytes of UTF-8 text in the buffer, excluding
	 *	         any terminating newline.
	 *	\sa      bytesInBuffer, charsInBuffer, charsInLine
	 */
	size_t bytesInLine() {
		size_t result = buffer.length - bufPos;
		if (result != 0 && buffer[$-1] == '\n') {
			return result - 1;
		} else {
			return result;
		}
	}

	//!	Returns the number of characters currently waiting in the internal
	//!	buffer.
	/*!
	 *	\return  the number of characters in the buffer, including the
	 *	         terminating newline if present.
	 *	\sa      bytesInBuffer, bytesInLine, charsInLine
	 */
	size_t charsInBuffer() {
		return toUCSindex(buffer[bufPos..$], buffer.length - bufPos);
	}

	//!	Returns the number of characters remaining on the current line of
	//!	input.
	/*!
	 *	\return  the number of characters in the buffer, excluding any
	 *	         terminating newline.
	 *	\sa      bytesInBuffer, charsInBuffer, charsInLine
	 */
	size_t charsInLine() {
		size_t result = charsInBuffer;
		if (result != 0 && buffer[$-1] == '\n') {
			return result - 1;
		} else {
			return result;
		}
	}

	//!	Reads a line of text from the input stream.
	/*!
	 *	\param withNewLine  whether to include the terminating newline
	 *	                    if it is present.  This makes a difference only
	 *	                    to the last line retrieved before EOF.
	 *	\return             the line retrieved, or \c null on EOF.
	 */
	char[] readLine(bool withNewLine = false) {
		if (eof) return null;

		scope(exit) bufPos = buffer.length = 0;

		if (!withNewLine && buffer[$-1] == '\n') {
			return buffer[bufPos..$-1].dup;
		} else {
			return buffer[bufPos..$].dup;
		}
	}

	//!	Determines whether EOF of the input stream has been reached.
	/*!
	 *	This function also reads a line of input into the buffer if it is
	 *	empty and there is any input to read.  Therefore, it will wait for
	 *	the user to input a line if one isn't already available.
	 *
	 *	\retval true   if the end of the input has been reached, and the
	 *	               buffer is empty.
	 *	\retval false  if there is input still to be read.
	 */
	bool eof() {
		if (bufPos != buffer.length) return false;
		if (_eof) return true;

		version (Windows) scope(exit) {
			char* bufz = toStringz(buffer);
			buffer = fromMBSz(bufz, 1);
		}

		buffer.length = 0;
		bufPos = 0;

		while (true) {
			int c = getc(stream);

			switch (c) {
				case '\r': case '\n':
					buffer ~= '\n';
					return false;

				case -1:
					_eof = 1;
					return buffer.length == 0;

				default:
					buffer ~= cast(char) c;
			}
		}
	}

	private {
		char[] buffer;
		size_t bufPos;
		FILE* stream = stdin;
		/*!	The EOF state of the input stream.  Values:
		 *	0 - not EOF
		 *	1 - EOF, line may be lost
		 *	2 - EOF, lost line recovered
		 */
		int _eof = 0;

		this() {}
	}

	invariant {
		assert (bufPos <= buffer.length);
		if (_eof) assert (buffer.length == 0);
	}
}


//!	The ConsoleInput class used to write out text to the standard
//!	output and error streams.
/*!
 *	This class buffers any incomplete UTF-8 characters that are passed into
 *	it.  This makes it suited to applications that transmit UTF-8 text one
 *	byte at a time, or in chunks whose boundaries may not coincide with
 *	character boundaries.
 */
class ConsoleOutput {
	//!	Writes out a single Unicode character.
	/*!
	 *	\param c  character to output.
	 *	\pre      There must be no partial UTF-8 characters remaining in the
	 *	          buffer.
	 */
	void write(dchar c)
	in {
		assert (bufPos == buffer.length);
	}
	body {
		buffer.length = bufPos = 0;
		encode(buffer, c);
		flushBuffer();
	}

	version (Windows) {
		//!	Writes out a string.
		/*!
		 *	\param str  the string to output.
		 */
		void write(char[] c) {
			buffer ~= c;
			flushBuffer();
		}

		//!	Writes out a string followed by a newline.
		/*!
		 *	\param str  the string to output.
		 *	\pre        For this to work correctly, c must not end partway
		 *	            through a UTF-8 character.
		 */
		void writeLine(char[] c) {
			buffer ~= c ~ '\n';
			flushBuffer();
		}

	} else {
		//!	Writes out a string.
		/*!
		 *	\param str  the string to output.
		 */
		void write(char[] str) {
			if (bufPos != buffer.length) {
				buffer ~= str;
				flushBuffer();
			} else {
				fwrite(str.ptr, 1, str.length, stream);
			}
		}

		//!	Writes out a string followed by a newline.
		/*!
		 *	\param str  the string to output.
		 *	\pre        For this to work correctly, c must not end partway
		 *	            through a UTF-8 character.
		 */
		void writeLine(char[] str) {
			if (bufPos != buffer.length) {
				buffer ~= str ~ '\n';
				flushBuffer();
			} else {
				fwrite(str.ptr, 1, str.length, stream);
				putc(stream, '\n');
			}
		}
	}

	//!	Performs formatted output on the stream.
	void writef(...) {
		return writefx(_arguments, _argptr);
	}

	//!	Performs formatted output on the stream.
	void writefln(...) {
		return writefx(_arguments, _argptr, true);
	}

	//!	Performs formatted output on the stream.
	void writefx(TypeInfo[] arguments, void* argptr, bool newline = false)
	in {
		assert (bufPos == buffer.length);
	} body {
		doFormat(delegate void(dchar c) { encode(buffer, c); },
		  arguments, argptr);
		if (newline) buffer ~= '\n';
		flushBuffer();
	}

	//!	Makes sure all pending output has actually been output.
	void flush() {
		flushBuffer();
		fflush(stream);
	}

	private {
		char[] buffer;
		size_t bufPos;
		bool _eof = false;
		FILE* stream;
		//!	Whether to fix the bug in Win9x whereby a line of output is lost
		//!	when stdin reaches EOF.
		/*!
		 *	\defaultValue  true for Win9x; false for NT/2000 and above.  If
		 *	               you discover that this is wrong for any Windows
		 *	               version from 95/NT4 upwards, please fix the
		 *	               processing of GetVersion accordingly.
		 */
		bool fixEofBug;

		this(FILE* s) {
			stream = s;

			version (Windows) {
				fixEofBug = (GetVersion() & 0x80000000)
				  && isatty(stdin._file) && isatty(s._file);
			}
		}

		version (Windows) {
			void flushBuffer() {
				version (Windows) {
					// first check if an extra line of output is needed
					if (fixEofBug && cIn._eof == 1) {
						cIn._eof = 2;
						putc('\n', stream);
					}
				}

				// find last whole character
				char expected;
				size_t newBufPos;

				foreach_reverse (index, char c; buffer) {
					debug if (this !is cErr) {
						cErr.writefln(
						  "Index %d, character %02x, expect %02x",
						  index, cast(ubyte) c, cast(ubyte) expected);
					}
					if (!(c & 0x80)) {
						newBufPos = index + 1;
						break;
					} else if (!(c & 0x40)) {
						if (expected == char.init) {
							expected = 0xC0;
						} else {
							expected = (expected >> 1) | 0x80;
						}
					} else {
						assert ((c & 0xC0) == 0xC0);
						if ((c & expected) == expected) {
							newBufPos = buffer.length;
						} else {
							newBufPos = index;
						}
						break;
					}
				}

				char[] toWrite = buffer[bufPos..newBufPos];
				fputs(toMBSz(toWrite, 1), stream);
				bufPos = newBufPos;

				if (bufPos == buffer.length) bufPos = buffer.length = 0;
			}

		} else {
			void flushBuffer() {
				fwrite(&buffer[bufPos], 1, buffer.length - bufPos, stream);
				bufPos = buffer.length = 0;
			}
		}
	}

	invariant {
		assert (bufPos <= buffer.length);
	}
}


//!	The standard input stream.
ConsoleInput cIn;

//!	The standard output stream.
ConsoleOutput cOut;

//!	The standard error stream.
ConsoleOutput cErr;


static this() {
	cIn = new ConsoleInput;
	cOut = new ConsoleOutput(stdout);
	cErr = new ConsoleOutput(stderr);
}
