/***********************************************************************\
*                              expression.d                             *
*                                                                       *
*                            LOLA interpreter                           *
*                                                                       *
*                              Version 0.2                              *
*                                                                       *
*                           Expression classes                          *
>***********************************************************************<
* Copyright (C) 2007 Stewart Gordon.                                    *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/lola/.                    *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.lola.expression;
import
	std.math2,
	std.string,
	smjg.libs.util.console;

abstract class Expression {
	public {
		abstract {
			bool isWHNF();
			Expression reduce();
			Expression reduceSingle();
			int freeVarLevel();
			Expression promoteFreeVars(int fromLevel);
			Expression substitute(Expression arg, int varLevel);
		}

		Expression reduceToWHNF()
		out (result) {
			assert (result.isWHNF);
		}
		body {
			Expression result = this;
			while (!result.isWHNF) result = result.reduce();
			return result;
		}

		Expression applyTo(Expression arg)
		in {
			debugOutput(format("Entered Expression.applyTo: %s ; %s", this, arg));
		}
		out (result) {
			debugOutput(format("Exiting Expression.applyTo: %s ; %s -> %s", this, arg, result));
		}
		body {
			return new Appl(this, arg);
		}

		Expression deref() { return this; }
	}
}


class ExpressionRef : Expression {
	public {
		static ExpressionRef opCall(Expression e) {
			ExpressionRef result = cast(ExpressionRef) e;

			if (result is null) {
				return new ExpressionRef(e);
			} else {
				return result.collapse;
			}
		}

		override {
			bool isWHNF() {
				return expression.isWHNF;
			}

			ExpressionRef reduce() {
				expression = expression.reduce();
				return collapse;
			}

			ExpressionRef reduceSingle() {
				expression = expression.reduceSingle();
				return collapse;
			}

			int freeVarLevel() { return expression.freeVarLevel; }

			ExpressionRef reduceToWHNF() {
				expression = expression.reduceToWHNF;
				return collapse;
			}

			Expression applyTo(Expression arg) {
				return expression.applyTo(arg);
			}

			Expression promoteFreeVars(int fromLevel) {
				if (freeVarLevel < fromLevel) {
					return this;
				} else {
					return expression.promoteFreeVars(fromLevel);
				}
			}

			Expression substitute(Expression arg, int varLevel) {
				if (expression.freeVarLevel < varLevel) {
					return this;
				} else {
					return expression.substitute(arg, varLevel);
				}
			}

			Expression deref() { return expression.deref; }

			char[] toString() {
				return expression.toString;
			}
		}
	}

	protected {
		this(Expression e) { expression = e; }

		ExpressionRef collapse() {
			if (cast(ExpressionRef) expression) {
				return (cast(ExpressionRef) expression).collapse;
			} else {
				return this;
			}
		}

		Expression expression;
	}
}


class NamedExpression : ExpressionRef {
	public {
		this(char name, Expression value) {
			super(value);
			_name = name;
		}

		override char[] toString() { return [_name]; }
	}

	private char _name;
}


class Abstraction : Expression {
	public {
		this(Expression b)
		in {
			debugOutput(format("Entered Abstraction.this: %s", b));
		}
		out {
			debugOutput(format("Exiting Abstraction.this: %s ; %d", b, level));
		}
		body {
			abstractionBody = ExpressionRef(b);

			level = b.freeVarLevel - 1;
		}

		override {
			bool isWHNF() { return true; }

			Abstraction reduce()
			in {
				debugOutput(format("Entered Abstraction.reduce: %s", this));
			}
			out {
				debugOutput(format("Exiting Abstraction.reduce: %s", this));
			}
			body {
				abstractionBody.reduce();
				return this;
			}

			Abstraction reduceSingle()
			in {
				debugOutput(format("Entered Abstraction.reduceSingle: %s", this));
			}
			out {
				debugOutput(format("Exiting Abstraction.reduceSingle: %s", this));
			}
			body {
				abstractionBody.reduceSingle();
				return this;
			}

			int freeVarLevel() { return level; }

			Expression applyTo(Expression arg)
			in {
				debugOutput(format("Entered Abstraction.applyTo: %s ; %s : %d", this, arg, level));
			}
			out (result) {
				debugOutput(format("Exiting Abstraction.applyTo: %s ; %s -> %s", this, arg, result));
			}
			body {
				if (level < 0) {
					return abstractionBody;
				} else {
					return abstractionBody.substitute(arg, 1);
				}
			}

			Expression promoteFreeVars(int fromLevel) {
				if (freeVarLevel < fromLevel) {
					return this;
				} else {
					return abstractionBody.promoteFreeVars(fromLevel + 1);
				}
			}

			Expression substitute(Expression arg, int varLevel)
			in {
				debugOutput(format("Entered Abstraction.substitute: %s ; %s : %d/%d", this, arg, level, varLevel));
			}
			out (result) {
				debugOutput(format("Exiting Abstraction.substitute: %s ; %s -> %s", this, arg, result));
			}
			body {
				if (level < varLevel) {
					return this;
				} else {
					return new Abstraction(abstractionBody.substitute(arg.promoteFreeVars(1), varLevel + 1));
				}
			}

			char[] toString() {
				return "(\\" ~ abstractionBody.toString ~ ')';
			}
		}
	}

	private {
		ExpressionRef abstractionBody;
		int level;
	}
}


class Appl : Expression {
	public {
		this(Expression o1, Expression o2)
		in {
			debugOutput(format("Entered Appl.this: %s ; %s", o1, o2));
		}
		out {
			debugOutput(format("Exiting Appl.this: %s ; %s : %d", o1, o2, level));
		}
		body {
			operator = ExpressionRef(o1);
			operand = ExpressionRef(o2);

			level = max(o1.freeVarLevel, o2.freeVarLevel);

		}

		override {
			bool isWHNF() { return false; }

			Expression reduce()
			in {
				debugOutput(format("Entered Appl.reduce, entering operator.reduceToWHNF: %s", this));
			}
			out {
				debugOutput(format("Exited operator.applyTo; exiting Appl.reduce: %s", this));
			}
			body {
				operator.reduceToWHNF();
				debugOutput(format("Exited operator.reduceToWHNF; entering operator.applyTo: %s", this));
				return operator.applyTo(operand);
			}

			Expression reduceSingle()
			in {
				debugOutput(format("Entered Appl.reduceSingle: %s", this));
			}
			out {
				debugOutput(format("Exiting Appl.reduceSingle: %s", this));
			}
			body {
				if (operator.isWHNF) {
					return operator.applyTo(operand);
				} else {
					operator.reduceSingle();
					return this;
				}
			}

			int freeVarLevel() { return level; }

			Expression promoteFreeVars(int fromLevel) {
				if (freeVarLevel < fromLevel) {
					return this;
				} else {
					return new Appl(operator.promoteFreeVars(fromLevel), operand.promoteFreeVars(fromLevel));
				}
			}

			Expression substitute(Expression arg, int varLevel)
			in {
				debugOutput(format("Entered Appl.substitute: %s ; %s : %d/%d", this, arg, level, varLevel));
			}
			out (result) {
				debugOutput(format("Exiting Appl.substitute: %s ; %s -> %s", this, arg, result));
			}
			body {
				if (level < varLevel) {
					return this;
				} else {
					return new Appl(operator.substitute(arg, varLevel),
					  operand.substitute(arg, varLevel));
				}
			}

			char[] toString() {
				return operator.toString ~ operand.toString ~ ',';
			}
		}
	}

	private {
		ExpressionRef operator, operand;
		int level;
	}
}


class Variable : Expression {
	public {
		static Variable opCall(int level)
		in {
			assert (level >= 1);
		}
		body {
			if (level >= pool.length) pool.length = level + 1;
			if (pool[level] is null) pool[level] = new Variable(level);
			return pool[level];
		}

		override {
			bool isWHNF() { return true; }
			Variable reduce() { return this; }
			Variable reduceSingle() { return this; }
			int freeVarLevel() { return level; }

			Expression promoteFreeVars(int fromLevel) {
				if (freeVarLevel < fromLevel) {
					return this;
				} else {
					return Variable(level + 1);
				}
			}

			Expression substitute(Expression arg, int varLevel)
			in {
				debugOutput(format("Entered Variable.substitute: %s ; %s : %d/%d", this, arg, level, varLevel));
			}
			out (result) {
				debugOutput(format("Exiting Variable.substitute: %s ; %s -> %s", this, arg, result));
			}
			body {
				if (level == varLevel) {
					return arg;
				} else if (level > varLevel) {
					return Variable(level - 1);
				} else {
					return this;
				}
			}

			char[] toString() {
				if (level < 26) {
					return ['a' - 1 + level];
				} else {
					return format("{%d}", level);
				}
			}
		}
	}

	private {
		static Variable[] pool;

		this(int l) {
			level = l;
		}

		int level;
	}
}

/+unittest {
	Variable v = Variable(1);
	Abstraction la = new Abstraction(v);
	ExpressionRef er = ExpressionRef(la);
	Appl a = new Appl(la, er);
}+/


//import std.stdio;

/+
void main() {
	Variable a = Variable(1);
	Abstraction la = new Abstraction(a);
	ExpressionRef er = ExpressionRef(la);
	Expression ap = new Appl(la, er);

	writefln(a);
	writefln(la);
	writefln(er);
	testReduce("Simple application test", ap);
}
+/


package void testReduce(char[] heading, inout Expression exp) {
	char[] s1 = exp.toString;
	char[] s2;

	cOut.writefln(heading);

	while (s1 != s2) {
		cOut.writefln(s1);
		s2 = s1;
		exp = exp.reduceSingle;
		s1 = exp.toString;
	}
	cOut.writefln();
}


debug {
	uint indentLevel = 0;

	void debugOutput(char[] text)
	in {
		assert (text.length >= 7);
		if (text[0..7] == "Exiting") assert (indentLevel != 0);
	}
	body {
		if (text[0..7] == "Exiting" || find(text, "exiting") != -1) indentLevel--;
		for (uint i = 0; i < indentLevel; i++) cOut.write(' ');
		cOut.writeLine(text);
		if (text[0..7] == "Entered") indentLevel++;
	}
} else {
	void debugOutput(lazy char[] text) {}
}
