/***********************************************************************\
*                               integer.d                               *
*                                                                       *
*                            LOLA interpreter                           *
*                                                                       *
*                              Version 0.2                              *
*                                                                       *
*    Integer arithmetic for internals and Church integer optimisation   *
>***********************************************************************<
* Copyright (C) 2007 Stewart Gordon.                                    *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/lola/.                    *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.lola.church;
import
	std.string,
	smjg.prog.lola.expression,
	smjg.prog.lola.exception;

debug import std.stdio;

class Integer : Expression {
	static this() {
		C_ZERO = Integer(0, true);
		C_ONE = Integer(1, true);
		ZERO = Integer(0, false);
	}

	public {
		static {
			Integer C_ZERO, C_ONE, ZERO;

			Integer opCall(uint v, bool c) {
				Integer[uint]* pool = c ? &churchPool : &plainPool;

				if (!(v in *pool)) {
					(*pool)[v] = new Integer(v, c);
				}
				return (*pool)[v];
			}

			Expression unchurch(Expression e) {
				Integer i = cast(Integer) e.deref;
				if (i is null) {
					return e.applyTo(Increment(false)).applyTo(ZERO).reduceToWHNF
					  .deref;

				} else {
					if (i._church) {
						return Integer(i.value, false);
					} else {
						throw new ProgramRuntimeException("Invalid Church integer");
					}
				}
			}
		}

		override {
			bool isWHNF() { return true; }
			Integer reduce() { return this; }
			Integer reduceSingle() { return this; }
			int freeVarLevel() { return _value == 0 ? -1 : 0; }
			Integer promoteFreeVars(int direction) { return this; }
			Integer substitute(Expression arg, int varLevel) { return this; }

			Expression applyTo(Expression arg) {
				if (!_church) throw new ProgramRuntimeException("Invalid Church integer");

				// TODO: optimise for applyTo(ChurchInteger) - exponentiation
				switch (_value) {
					case 0:
						return C_ONE;

					case 1:
						return arg;

					default:
						return new IteratedFunction(_value, arg);
				}
			}

			char[] toString() {
				return format(_church ? "{%d}" : "[%d]", _value);
			}
		}

		uint value() { return _value; }
		bool church() { return _church; }
	}

	private {
		static Integer[uint] churchPool, plainPool;

		this(uint v, bool c) {
			_value = v;
			_church = c;
		}

		uint _value;
		bool _church;
	}
}


class Increment : Expression {
	static this() {
		ba = new Appl(Variable(2), Variable(1));
		_churchInstance = new Increment;
		_churchInstance._church = true;
		_plainInstance = new Increment;
	}

	public {
		static Increment opCall(bool c) {
			return c ? _churchInstance : _plainInstance;
		}

		override {
			bool isWHNF() { return true; }
			Increment reduce() { return this; }
			Increment reduceSingle() { return this; }
			int freeVarLevel() { return 0; }
			Increment promoteFreeVars(int fromLevel) { return this; }
			Increment substitute(Expression arg, int varLevel) { return this; }

			Expression applyTo(Expression e) {
				/*
				 *	Operand >  church  plain  other
				 *	this v
				 *	church     church  fail   rewrite as lambda
				 *	plain      fail    plain  reduce and retry
 				 *
				 *	Church increment applied to plain integer
				 *	\\\cb,ba,, [n]
				 *	\\[n]b,ba,,
				 *	can reduce, but is it any use?
				 */

				Integer i = cast(Integer) e.deref;

				if (i !is null && _church == i.church) {
					if (i.value == uint.max) {
						if (!_church) {
							throw new ProgramRuntimeException("Overflow of output value");
						} // otherwise, fall through to lambda case
					} else {
						return Integer(i.value + 1, _church);
					}
				} else if (_church) {
					/*	rewrite as a lambda abstraction
					 *	+ is
					 *		\\\cb,ba,,
					 *	so +X, is
					 *		\\Xb,ba,,
					 */
					return new Appl(new Appl(e, Variable(2)), ba);
				} else {
					// reduce and retry
					if (e.isWHNF) {
						throw new ProgramRuntimeException("Invalid Church integer");
					} else {
						return applyTo(e.reduceToWHNF);
					}
				}
			}

			char[] toString() {
				return _church ? "{+}" : "[+]";
			}
		}

		bool church() { return _church; }
	}

	private {
		this() {}

		bool _church;

		static {
			Appl ba;
			Increment _churchInstance, _plainInstance;
		}
	}
}


class IteratedFunction : Expression {
	public {
		this(uint c, Expression f) {
			_count = c;
			_func = ExpressionRef(f);
		}

		override {
			bool isWHNF() { return _func.isWHNF; }
			IteratedFunction reduce() { _func.reduce(); return this; }
			Integer reduceSingle() { _func.reduceSingle(); return this; }
			int freeVarLevel() { return _func.freeVarLevel; }

			IteratedFunction promoteFreeVars(int fromLevel) {
				if (freeVarLevel < fromLevel) {
					return this;
				} else {
					return new IteratedFunction(_count,
					  _func.promoteFreeVars(fromLevel));
				}
			}

			IteratedFunction substitute(Expression arg, int varLevel) {
				if (_func.freeVarLevel < varLevel) {
					return this;
				} else {
					return new IteratedFunction(_count,
					  _func.substitute(arg, varLevel));
				}
			}

			Expression applyTo(Expression arg) {
				for (uint i = 0; i < _count; i++) {
					arg = _func.applyTo(arg);
				}
				return arg;
			}

			char[] toString() {
				return format("{%d %s}", _count, _func);
			}
		}
	}

	private {
		uint _count;
		ExpressionRef _func;
	}
}


unittest {
	Variable a = Variable(1);
	Variable b = Variable(2);
	Variable c = Variable(3);

	Expression plus = new Abstraction(new Abstraction(new Abstraction(
		new Appl(new Appl(c, b), new Appl(b, a)))));
	testReduce("Plus operator", plus);

	Expression times = new Abstraction(new Abstraction(
		new Appl(
			new Appl(b, new Appl(a, plus)),
			new Abstraction(new Abstraction(a))
		)
	));

	Expression i2 = new Abstraction(new Abstraction(new Appl(b, new Appl(b, a))));
	testReduce("Church integer 2", i2);

	Expression c2 = Integer.unchurch(i2);
	testReduce("Unchurching...", c2);

	Expression i3 = new Appl(plus, i2);
	testReduce("Church integer 3", i3);

	Expression c3 = Integer.unchurch(i3);
	testReduce("Unchurching...", c3);

	Expression i4 = new Appl(i2, i2);
	testReduce("Church integer 4", i4);

	Expression c4 = Integer.unchurch(i4);
	testReduce("Unchurching...", c4);

	Expression i5 = new Appl(plus, i4);
	testReduce("Church integer 5", i5);

	Expression c5 = Integer.unchurch(i5);
	testReduce("Unchurching...", c5);

	Expression i6 = new Appl(new Appl(times, i2), i3);
	testReduce("Church integer 6", i6);

	Expression c6 = Integer.unchurch(i6);
	testReduce("Unchurching...", c6);

	Expression i8 = new Appl(i3, i2);
	testReduce("Church integer 8", i8);

	Expression c8 = Integer.unchurch(i8);
	testReduce("Unchurching...", c8);

	Expression i10 = new Appl(new Appl(times, i2), i5);
	testReduce("Church integer 10", i10);

	Expression c10 = Integer.unchurch(i10);
	testReduce("Unchurching...", c10);

	Expression i12 = new Appl(new Appl(times, i2), i6);
	testReduce("Church integer 12", i12);

	Expression c12 = Integer.unchurch(i12);
	testReduce("Unchurching...", c12);
}
