/***********************************************************************\
*                                parser.d                               *
*                                                                       *
*                            LOLA interpreter                           *
*                                                                       *
*                              Version 0.2                              *
*                                                                       *
*                           Parsing algorithm                           *
>***********************************************************************<
* Copyright (C) 2007 Stewart Gordon.                                    *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/lola/.                    *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.lola.parser;
import
	smjg.prog.lola.expression,
	smjg.prog.lola.exception,
	std.string;
debug import std.stdio;

private class PendingNamedExpression : Expression {
	this(char n) { name = n; }

	override {
		bool isWHNF() { return false; }

		Expression reduce() { return symbolTable[name]; }
		Expression reduceSingle() { return symbolTable[name]; }

		char[] toString() { return [name]; }
		int freeVarLevel() { return 0; }
		Expression promoteFreeVars(int fromLevel) { return this; }
		Expression substitute(Expression arg, int varLevel) { return this; }
	}

	char name;
}


private Expression[char] symbolTable;


Expression parse(char[] code) {
	char[][] lines = splitlines(code);
	Expression mainExpression;

	foreach (iLine, line; lines) {
		Expression[] expStack;
		size_t[] lambdaStack;

	parseLine:
		foreach (c; line) {
			switch (c) {
				case ' ': case '\t':
					break parseLine;

				case '\\':
					lambdaStack ~= expStack.length;
					break;

				case ',':
					if (expStack.length < 2) {
						throw new ProgramParseException(
							"',' encountered without sufficient preceding subexpressions", iLine, line);
					}
					if (lambdaStack.length != 0 && lambdaStack[$-1] == expStack.length) {
						throw new ProgramParseException(`',' encountered immediately after '\'`, iLine, line);
					}

					// first check for lambda abstractions to be closed
					while (lambdaStack.length != 0 && lambdaStack[$-1] == expStack.length - 1) {
						expStack[$-1] = new Abstraction(expStack[$-1]);
						lambdaStack.length = lambdaStack.length - 1;
					}

					// now join the top two subexpressions together
					expStack[$-2] = new Appl(expStack[$-2], expStack[$-1]);
					expStack.length = expStack.length - 1;
					break;

				default:
					if (c >= 'a' && c <= 'z' && c - 'a' < lambdaStack.length) {
						expStack ~= Variable(c - 'a' + 1);
					} else {
						if (!(c in symbolTable)) {
							symbolTable[c] = new PendingNamedExpression(c);
						}
						expStack ~= symbolTable[c];
					}
			}
		}
		// close any pending lambda abstractions
		while (lambdaStack.length != 0 && lambdaStack[$-1] == expStack.length - 1) {
			expStack[$-1] = new Abstraction(expStack[$-1]);
			lambdaStack.length = lambdaStack.length - 1;
		}

		switch (expStack.length) {
			case 0:  // no code on the line
				break;

			case 1:  // main function
				assert (lambdaStack.length == 0);
				if (mainExpression !is null) {
					throw new ProgramParseException("Duplicate main function", iLine, line);
				}
				assert (expStack[0].freeVarLevel <= 0);
				mainExpression = expStack[0];
				break;

			case 2:  // named function
				PendingNamedExpression pnf = cast(PendingNamedExpression) expStack[0];

				assert (expStack[1].freeVarLevel <= 0);
				if (lambdaStack.length != 0) {
					throw new ProgramParseException(
						"Too many expressions - check for missing ','", iLine, line);
				}

				if (pnf is null) {
					if (cast(NamedExpression) expStack[0]) {
						throw new ProgramParseException("Duplicate named function " ~ expStack[0].toString, iLine, line);
					} else {
						throw new ProgramParseException(
							"Too many expressions - check for missing ','", iLine, line);
					}
				}

				symbolTable[pnf.name] = new NamedExpression(pnf.name, expStack[1]);
				break;

			default:
				throw new ProgramParseException(
					"Too many expressions - check for missing ','", iLine, line);
		}
	}

	if (mainExpression is null) {
		throw new ProgramParseException("No main expression");
	}

	debug writefln("\nSymbol table:");
	foreach (name, exp; symbolTable) {
		if (cast(PendingNamedExpression) exp) {
			throw new ProgramParseException("Undefined symbol '" ~ name ~ "'");
		}

		debug writefln("%s : %s", name, exp.deref);
	}
	debug getch();

	return mainExpression;
}
